/*
 * Copyright (c) 2017-2022, StrategyQuant - All rights reserved.
 *
 * Code in this file was made in a good faith that it is correct and does what it should.
 * If you found a bug in this code OR you have an improvement suggestion OR you want to include
 * your own code snippet into our standard library please contact us at:
 * https://roadmap.strategyquant.com
 *
 * This code can be used only within StrategyQuant products.
 * Every owner of valid (free, trial or commercial) license of any StrategyQuant product
 * is allowed to freely use, copy, modify or make derivative work of this code without limitations,
 * to be used in all StrategyQuant products and share his/her modifications or derivative work
 * with the StrategyQuant community.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
 * INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 */
package com.strategyquant.plugin.Servlet.impl.Project;

import com.strategyquant.pluginlib.annotations.Category;
import com.strategyquant.pluginlib.annotations.License;
import com.strategyquant.pluginlib.annotations.Name;
import com.strategyquant.pluginlib.annotations.ShortDesc;
import com.strategyquant.tradinglib.servlet.IServletPlugin;
import com.strategyquant.webguilib.servlet.HttpJSONServlet;
import com.strategyquant.lib.*;
import com.strategyquant.lib.constants.*;
import com.strategyquant.tradinglib.correlation.*;
import com.strategyquant.tradinglib.project.*;
import com.strategyquant.datalib.*;
import com.strategyquant.tradinglib.*;
import com.strategyquant.tradinglib.results.*;
import com.strategyquant.tradinglib.results.SpecialValues;

import SQ.CorrelationOf.*;

import net.xeoh.plugins.base.annotations.PluginImplementation;
import net.xeoh.plugins.base.annotations.meta.Author;
import org.eclipse.jetty.server.Handler;
import org.eclipse.jetty.servlet.ServletContextHandler;
import org.eclipse.jetty.servlet.ServletHolder;
import org.json.JSONObject;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Map;
import java.util.ArrayList;
import java.util.Iterator;

@Author(name = "Tomas Brynda")
@Name(name = "Filter by correlation Servlet Plugin")
@Category(name = "Server Extension")
@License(text = "")
@ShortDesc(text = "")
@PluginImplementation
public class FilterByCorrelationServlet implements IServletPlugin {

    private static final Logger Log = LoggerFactory.getLogger(FilterByCorrelationServlet.class);
    private ServletContextHandler connectionContext;

    //------------------------------------------------------------------------
    //------------------------------------------------------------------------
    //------------------------------------------------------------------------

    @Override
    public String getProduct() {
        return SQConst.CODE_SQ;
    }

    //------------------------------------------------------------------------

    @Override
    public int getPreferredPosition() {
        return 0;
    }

    //------------------------------------------------------------------------

    @Override
    public void initPlugin() throws Exception {
    }

    //------------------------------------------------------------------------

    @Override
    public Handler getHandler() {
        if(connectionContext == null){
            connectionContext = new ServletContextHandler(ServletContextHandler.SESSIONS);
            connectionContext.setContextPath("/filterByCorrelation/");
            connectionContext.addServlet(new ServletHolder(new FilterByCorrelation()),"/*");
        }
        return connectionContext;
    }

    //------------------------------------------------------------------------

    class FilterByCorrelation extends HttpJSONServlet {
        @Override
        protected String execute(String command, Map<String, String[]> args, String method){
            switch(command) {
                case "filter": return onFilter(args);
                default:
                    return apiErrorJSON("Execution failed. Unknown command '"+command+"'.", null);
            }
        }

        //------------------------------------------------------------------------

        private String onFilter(Map<String, String[]> args) {
            JSONObject response = new JSONObject();
            try {
                String projectName = tryGetParam(args, "projectName")[0];
                String databankName = tryGetParam(args, "databankName")[0];
                int period = Integer.parseInt(tryGetParam(args, "period")[0]);
                double maxCorrelation = Double.parseDouble(tryGetParam(args, "maxCorrelation")[0]);

                int removedCount = filterByCorrelation(projectName, databankName, period, maxCorrelation);

                response.put("success", "Strategies filtered, "+removedCount+" strategies removed.");
                return response.toString();
            }
            catch(Exception e){
                Log.error("Error while filtering strategies by correlation", e);
                response.put("error", "Filtering failed - " + e.getMessage());
                return response.toString();
            }
        }

        //------------------------------------------------------------------------

        private int filterByCorrelation(String projectName, String databankName, int period, double maxCorrelation) throws Exception {
            Log.info("Filtering by correlation (Project: " + projectName + ", databank: " + databankName + ")...");
            SQProject project = ProjectEngine.get(projectName);
            Databank databank = project.getDatabanks().get(databankName);
            int removedCount = 0;

            // Check for correlated strategies using fitness score as a ranking/priority list
            ArrayList<ResultsGroup> rgSortedCandidates = new ArrayList<>(databank.getRecords());

            rgSortedCandidates.sort((o1, o2) -> Double.valueOf(o1.getFitness()).compareTo(Double.valueOf(o2.getFitness())));

            ArrayList<ResultsGroup> rgSortedCandidatesRanked = new ArrayList();

            for(ResultsGroup rgSC : rgSortedCandidates) {
                rgSortedCandidatesRanked.add(0, rgSC);
            }

            ArrayList<ResultsGroup> rgUncorrelated = new ArrayList();
            // todo: multiple strategies could have the same fitness. this should be addressed using additional sorting (i.e. fitness > pf > ret/dd .. etc)

            while (rgSortedCandidatesRanked.size() > 0) {
                ResultsGroup focusStrat = rgSortedCandidatesRanked.get(0).clone();
                String focusStratName = focusStrat.getName();

                Log.info("Focus strategy to compare: {}", focusStratName);

                rgUncorrelated.add(focusStrat.clone()); // add the best strategy in the pool to the uncorrelated list as a starting point
                rgSortedCandidatesRanked.remove(0);

                Iterator<ResultsGroup> it = rgSortedCandidatesRanked.iterator();

                while(it.hasNext()) {
                    ResultsGroup subStrat = it.next();

                    String subStratName = subStrat.getName();

                    if (subStratName != focusStratName) {
                        CorrelationPeriods corrPeriods = new CorrelationPeriods();

                        CorrelationType corrType = CorrelationTypes.getInstance().findClassByName("ProfitLoss");
                        CorrelationComputer correlationComputer = new CorrelationComputer();

                        // compute correlation periods
                        CorrelationPeriods correlationPeriods = precomputePeriodsAP(focusStrat, subStrat, period, corrType);
                        double corrValue = SQUtils.round2(correlationComputer.computeCorrelation(false, focusStratName, subStratName, focusStrat.orders(), subStrat.orders(), correlationPeriods));

                        Log.info(" - Correlation of '{}'' to focus strategy: {}", subStratName, corrValue);

                        if (corrValue > maxCorrelation) {
                            it.remove();
                            removedCount++;
                            Log.info("   - Removing strategy " + subStratName + " (correlation " + corrValue + " > " + maxCorrelation + ")...");
                            databank.remove(subStratName, true, true, true, true, true, null);
                        }
                    } else {
                        it.remove();
                    }
                }
            }

            return removedCount;
        }

        //------------------------------------------------------------------------
        
        private CorrelationPeriods precomputePeriodsAP(ResultsGroup paramResultsGroup1, ResultsGroup paramResultsGroup2, int paramInt, CorrelationType paramCorrelationType) throws Exception {
            CorrelationPeriods correlationPeriods = new CorrelationPeriods();
            TimePeriod timePeriod = CorrelationLib.getPeriod(paramResultsGroup1.orders());
            long l1 = timePeriod.from;
            long l2 = timePeriod.to;
            timePeriod = CorrelationLib.getPeriod(paramResultsGroup2.orders());
            if (timePeriod.from < l1) {
                l1 = timePeriod.from;
            }
            if (timePeriod.to > l2) {
                l2 = timePeriod.to;
            }

            TimePeriods timePeriods1 = CorrelationLib.generatePeriods(paramInt, l1, l2);
            TimePeriods timePeriods2 = timePeriods1.clone();

            paramCorrelationType.computePeriods(paramResultsGroup1.orders(), paramInt, timePeriods1);
            correlationPeriods.put(paramResultsGroup1.getName(), timePeriods1);

            paramCorrelationType.computePeriods(paramResultsGroup2.orders(), paramInt, timePeriods2);
            correlationPeriods.put(paramResultsGroup2.getName(), timePeriods2);

            return correlationPeriods;
        }        
    }


}
